"""Provides Additional Extensions for Snowpark"""


from .dataframe_extensions import *
from .functions_extensions import *
from .session_builder_extensions import *
from .types_extensions import *
from .column_extensions import *


from snowflake.snowpark import Column
from snowflake.snowpark.functions import get, lit

if not hasattr(Column,"___extended"):
    Column.__extended = True
    def getItem(self, index):
        name = str(self._expression).replace("\"",'') + f"[{index}]"
        return get(self,lit(index)).alias(name)

    Column.getItem = getItem




from snowflake.snowpark.types import StructType, StructField

if not hasattr(StructType, "___extended"):
  StructType.___extended = True    
  def schema__get_item__(self,name):
      name = name.upper()
      for f in self.fields:
         if name == f.name.upper():
            return f
  StructType.__getitem__ = schema__get_item__
  StructField.dtype = property(lambda self: self.datatype)


import logging
import configparser
from pathlib import Path
from snowflake.snowpark import Session
import os
import shortuuid

if not hasattr(Session.SessionBuilder,"___extended"):

    Session.SessionBuilder.___extended = True
    Session.SessionBuilder.__old_create = Session.SessionBuilder.create
    def SessionBuilder_updated_create(self):
        session = self.__old_create()
        if hasattr(self,"__appname__"):
            setattr(session, "__appname__", self.__appname__)
            uuid = shortuuid.uuid()
            session.query_tag = f"APPNAME={session.__appname__};execution_id={uuid}"
        return session
    Session.SessionBuilder.create = SessionBuilder_updated_create
    def SessionBuilder_appName(self,name):
        self.__appname__ = name
        return self

    Session.SessionBuilder.appName = SessionBuilder_appName
    def append_tag(self,tag:str):
        session.query_tag = session.query_tag + ";" + tag
    Session.append_tag = append_tag    

    def SessionBuilder_getOrCreate(self):
        from snowflake.snowpark import context
        from snowflake.snowpark.session import _session_management_lock, _active_sessions
        with _session_management_lock:
            if len(_active_sessions) == 1:
                return next(iter(context._active_sessions))
        return self.create()


    Session.SessionBuilder.getOrCreate = SessionBuilder_getOrCreate


    def SessionBuilder_env(self):
        self._options["user"]     = os.getenv("SNOW_USER") or os.getenv("SNOWSQL_USER")
        self._options["password"] = os.getenv("SNOW_PASSWORD") or os.getenv("SNOWSQL_PWD")
        self._options["account"]  = os.getenv("SNOW_ACCOUNT") or os.getenv("SNOWSQL_ACCOUNT")
        self._options["role"]     = os.getenv("SNOW_ROLE") or os.getenv("SNOWSQL_ROLE")
        self._options["warehouse"]= os.getenv("SNOW_WAREHOUSE") or os.getenv("SNOWSQL_WAREHOUSE")
        self._options["database"] = os.getenv("SNOW_DATABASE") or os.getenv("SNOWSQL_DATABASE")
        return self

    def SessionBuilder_snowsql_config(self,section=None,configpath=f'{Path.home()}/.snowsql/config'):
        config = configparser.ConfigParser()
        if not os.path.exists(configpath):
            raise Exception(f"No config fount at:{configpath}")
        config.read(configpath)
        section = config["connections" if section is None else f"connections.{section}"]
        connection_parameters = {
            "user":section.get("username"),
            "password":section.get("password"),
            "account":section.get('accountname'),
            "role":section.get('rolename'),
            "warehouse":section.get('warehousename'),
            "database":section.get("dbname"),
            "schema":section.get("schemaname")
        }
        return self.configs(connection_parameters)

    Session.SessionBuilder.from_snowsql = SessionBuilder_snowsql_config

    Session.SessionBuilder.env = SessionBuilder_env


from snowflake.snowpark.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    DataType,
    DateType,
    DecimalType,
    DoubleType,
    GeographyType,
    LongType,
    MapType,
    StringType,
    StructType,
    TimestampType,
    TimeType,
    VariantType,
)
import datetime
import decimal
# based on https://docs.snowflake.com/en/developer-guide/udf/python/udf-python-designing.html#label-sql-python-data-type-mappings

def map_to_python_type(dtype:DataType):
    if isinstance(dtype, ArrayType):
        return list
    elif isinstance(dtype, BinaryType):
        return bytes
    elif isinstance(dtype, BooleanType):
        return bool
    elif isinstance(dtype, DateType):
        return datetime.date
    elif isinstance(dtype, DecimalType):
        return int if dtype.scale == 0 else decimal.Decimal
    elif isinstance(dtype, DoubleType):
        return float
    elif isinstance(dtype, GeographyType):
        return dict
    elif isinstance(dtype, LongType):
        return int
    elif isinstance(dtype, MapType):
        return dict
    elif isinstance(dtype, StringType):
        return str
    elif isinstance(dtype, StructType):
        return dict
    elif isinstance(dtype, TimestampType):
        return datetime.datetime
    elif isinstance(dtype, TimeType):
        return datetime.time
    elif isinstance(dtype, VariantType):
        return dict



from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

VERSION = '0.0.1'

setup(name='snowpark_extensions',
      version=VERSION,
      description='A set of helpers to extend snowpark functionality',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/MobilizeNet/snowpark-extensions-py',
      author='mauricio.rojas',
      install_requires=['snowflake-snowpark-python','pandas','shortuuid'],
      author_email='mauricio.rojas@mobilize.net',
      packages=['snowpark_extensions'],
      zip_safe=False)



from snowflake.snowpark import DataFrame, Row
from snowflake.snowpark.functions import col, lit, udtf
from snowflake.snowpark import functions as F
import pandas as pd
import numpy as np
from snowpark_extensions.utils import map_to_python_type
import shortuuid
from snowflake.snowpark.types import StructType,StructField

if not hasattr(DataFrame,"___extended"):
    DataFrame.___extended = True
    def get_dtypes(schema):
        data = np.array([map_to_python_type(x.datatype) for x in schema.fields])
        # providing an index
        dtypes_series = pd.Series(data, index=schema.names)
        return dtypes_series

    DataFrame.dtypes = property(lambda self: get_dtypes(self.schema))


    def map(self,func,output_types,input_types=None,input_cols=None,to_row=False):
        clazz="map"+shortuuid.uuid()[:8]
        output_schema=[]
        if not input_types:
            input_types = [x.datatype for x in self.schema.fields]
        input_cols_len=len(input_types)
        if not input_cols:
            input_col_names=self.columns[:input_cols_len] 
            _input_cols = [self[x] for x in input_col_names]
        else:
            input_col_names=input_cols
            _input_cols = input_cols
        output_schema = [StructField(f"c_{i+1}",output_types[i]) for i in range(len(output_types))]
        output_cols = [col(x.name) for x in output_schema]
        if to_row:
            myRow = Row(*input_col_names)
            def process_func(self,*argv):
                yield func(myRow(*argv))
        else:
            def process_func(self,*argv):
                yield func(*argv)
        udtf_class = type(clazz, (object, ), {"process":process_func})
        udtf(udtf_class,output_schema=StructType(output_schema), input_types=input_types,name=clazz,replace=True,packages=["snowflake-snowpark-python"])
        return self.join_table_function(clazz,*_input_cols).select(*output_cols)

    DataFrame.map = map

    def simple_map(self,func):
        return self.select(*func(self))

    DataFrame.simple_map = simple_map    

    # EXPLODE HELPERS
    class Explode:
        def __init__(self,expr):
            self.expr = expr

    def explode(expr):
        return Explode(expr)

    F.explode = explode

    DataFrame.oldwithColumn = DataFrame.withColumn
    def withColumnExtended(self,colname,expr):
        if isinstance(expr, Explode):
            return self.join_table_function('flatten',date_range_udf(col("epoch_min"), col("epoch_max"))).drop(["SEQ","KEY","PATH","INDEX","THIS"]).rename("VALUE",colname)
        else:
            self.oldwithColumn(colname,expr)
    DataFrame.withColumn = withColumnExtended


import shortuuid
from snowflake.snowpark import Window, Column
from snowflake.snowpark.types import *
from snowflake.snowpark.functions import udtf, col
from snowflake.snowpark.relational_grouped_dataframe import RelationalGroupedDataFrame

if not hasattr(RelationalGroupedDataFrame, "applyInPandas"):
  def applyInPandas(self,func,output_schema):
      from snowflake.snowpark.functions import col
      input_types = [x.datatype for x in self._df.schema.fields]
      input_cols  = [x.name for x in self._df.schema.fields]
      output_cols = [x.name for x in output_schema.fields]
      grouping_exprs = [Column(x) for x in self._grouping_exprs]
      clazz="applyInPandas"+shortuuid.uuid()[:8]
      def __init__(self):
          self.vals = []
      def process(self, *val):
          self.vals.append(val)
          yield None
      def end_partition(self):
          import pandas as pd
          pandas_input = pd.DataFrame(self.vals, columns=input_cols)
          pandas_output = func(pandas_input)
          for row in pandas_output.itertuples(index=False):
             yield tuple(row)
      non_ambigous_output_schema = StructType([StructField(f"pd_{i}",output_schema.fields[i].datatype) for i in range(len(output_schema.fields))])
      renamed_back = [col(f"pd_{i}").alias(output_schema.fields[i].name) for i in range(len(non_ambigous_output_schema.fields))]
      udtf_class = type(clazz, (object, ), {"__init__":__init__,"process":process,"end_partition":end_partition})
      tfunc = udtf(udtf_class,output_schema=non_ambigous_output_schema, input_types=input_types,name=clazz,replace=True,is_permanent=False,packages=["snowflake-snowpark-python", "pandas"])
      return self._df.join_table_function(tfunc(*input_cols).over(partition_by=grouping_exprs, order_by=grouping_exprs)).select(*renamed_back)

  RelationalGroupedDataFrame.applyInPandas = applyInPandas
  ###### HELPER END   






from snowflake.snowpark import functions as F
from snowflake.snowpark import DataFrame

if not hasattr(F,"___extended"):
    F.___extended = True
    def unix_timestamp(col):
        return call_builtin("DATE_PART","epoch_second",col)

    def from_unixtime(col):
        return F.to_timestamp(col).alias('ts')

    def format_number(col,d):
        return F.to_varchar(col,'999,999,999,999,999.' + '0'*d)

    def reverse(col):
        return F.call_builtin('reverse',col)

    def date_add(col,num_of_days):
        return dateadd(lit('day'),col,num_of_days)

    def date_sub(col,num_of_days):
        return dateadd(lit('day'),col,-1 * num_of_days)

    def create_map(*col_names):
        """
        Usage:
        res = df.select(create_map('name', 'age').alias("map")).collect()
        """
        from snowflake.snowpark.functions import col,lit, object_construct
        col_list = []
        if isinstance(col_names, tuple):
            col_names = list(col_names[0])
        if isinstance(c, str):
            col_list.append(lit(c))
            col_list.append(col(c))
        else:
            col_list.append(lit(str(c._expression).replace("\"",'')))
            col_list.append(c)
        return object_construct(*col_list)

    def array_sort(array:list)->list:
        def compare(item1, item2=None):
            if item1 is None:
                return 1
            elif item2 is None:
                return -1
            elif item1 < item2:
                return -1
            elif item1 > item2:
                return 1
            else:
                return 0
        array.sort(key=compare)
        return array

    array_sort_udf=None
    def array_sort(array):
        if not array_sort_udf:
            array_sort_udf = udf(array_sort,return_type=ArrayType(),input_types=[ArrayType()],name="array_sort",is_permanent=False,replace=True)
        array_sort_udf(array)

    F.array_sort = array_sort
    F.create_map = create_map
    F.unix_timestamp = unix_timestamp
    F.from_unixtime = from_unixtime
    F.format_number = format_number
    F.reverse = reverse
    F.date_data = date_add
    F.date_sub = date_sub